from meeshkan.schemabuilder.servers import normalize_path_if_matches
from openapi_typed import Server
from typing import List
from hamcrest import *

from http_types import RequestBuilder

petstore_req = RequestBuilder.from_url(
    "https://petstore.swagger.io/v1/pets")

petstore_server = Server(url="https://petstore.swagger.io/v1")


def test_normalize_path_for_match():
    norm_pathname = normalize_path_if_matches(petstore_req, [petstore_server])
    assert_that(norm_pathname, is_("/pets"))


def test_no_match():
    req_no_match = RequestBuilder.from_url(
        "https://petstore.swagger.io/v2/pets")
    norm_pathname = normalize_path_if_matches(req_no_match, [petstore_server])
    assert_that(norm_pathname, is_(None))



import os
from pathlib import Path
from typing import Dict, Any, List
import yaml
import logging


PACKAGE_PATH = Path(os.path.dirname(os.path.realpath(__file__)))  # type: Path

LOG_CONFIG_FILE = PACKAGE_PATH.joinpath('logging.yaml')

BASE_DIR = Path.home().joinpath('.meeshkan')
LOGS_DIR = BASE_DIR.joinpath('logs')


# Don't automatically expose anything to top level, as the entire module is loaded as-is
__all__ = []  # type: List[str]

_SETUP_DONE = False


def _ensure_base_dirs(verbose=True):
    def create_dir_if_not_exist(path: Path):
        if not path.is_dir():
            # Print instead of logging as loggers may not have been configured yet
            if verbose:
                print("Creating directory {path}".format(path=path))
            path.mkdir()

    create_dir_if_not_exist(BASE_DIR)
    create_dir_if_not_exist(LOGS_DIR)


def _setup_logging(log_config: Path = LOG_CONFIG_FILE, silent: bool = False) -> None:
    """Setup logging configuration. This must be done before creating loggers.

    Keyword Arguments:
        log_config {Path} -- Path to logging configuration YAML (default: {LOG_CONFIG_FILE})
        silent {bool} -- Whether to remove non-file logging handlers or not. (default: {False})

    """

    if not log_config.is_file():
        raise RuntimeError(
            "Logging file {log_file} not found".format(log_file=log_config))

    with log_config.open() as log_file:
        config_orig = yaml.safe_load(log_file.read())  # type: Any

    def prepare_filenames(config: Dict[str, Any]) -> Dict[str, Any]:
        """
        Prepend `LOGS_DIR` to all 'filename' attributes listed for handlers in logging.yaml
        :param config: Configuration dictionary
        :return: Configuration with 'filename's prepended with LOGS_DIR
        """
        for handler_name in config['handlers'].keys():
            handler_config = config['handlers'][handler_name]
            if 'filename' in handler_config:
                filename = Path(handler_config['filename']).name
                handler_config['filename'] = str(LOGS_DIR.joinpath(filename))
        return config

    config = prepare_filenames(config_orig)
    logging.config.dictConfig(config)
    if silent:
        _remove_non_file_handlers()


def _remove_non_file_handlers():
    root_logger = logging.getLogger()  # Root logger
    for handler in root_logger.handlers.copy():
        if not isinstance(handler, logging.FileHandler):
            root_logger.handlers.remove(handler)


def setup(verbose=True):
    """
    Setup Meeshkan CLI by ensuring necessary directories exist and loggers have been configured.

    Keyword Arguments:
        verbose {bool} -- Verbose mode or not (default: {True})
    """
    global _SETUP_DONE
    if _SETUP_DONE:
        return
    _ensure_base_dirs(verbose=False)
    _setup_logging()
    _SETUP_DONE = True



from meeshkan.sources.kafka import KafkaProcessor, KafkaProcessorConfig, KafkaSource
from os import read
import pytest
from ..util import read_recordings_as_request_response
from hamcrest import *

exchanges = read_recordings_as_request_response()

kafkaProcessorConfig = KafkaProcessorConfig(
    broker="localhost:9092", topic="example")


@pytest.fixture()
def test_processor(event_loop):
    processor = KafkaProcessor(kafkaProcessorConfig)
    app = processor.app
    """passing in event_loop helps avoid 'attached to a different loop' error"""
    app.finalize()
    app.conf.store = 'memory://'
    app.flow_control.resume()
    return processor


@pytest.mark.asyncio()
async def test_processing(test_processor: KafkaProcessor):
    async with test_processor.agent.test_context() as agent:
        for exchange in exchanges:
            await agent.put(exchange)
        res = agent.results
        assert_that(list(res.keys()), has_length(168))
        assert_that(res[0], is_(exchanges[0]))


@pytest.mark.asyncio()
async def test_source(test_processor: KafkaProcessor):
    # TODO How to test the source stream with `agent.test_context()` (without running Kafka)?
    pass



import logging

from http_types import Request, Response

from .channel import Channel
from .proxy_callback import ProxyCallback

logger = logging.getLogger(__name__)


class ProxyBase(ProxyCallback):
    def __init__(self, data_callback, router, ssl_options=None):
        super().__init__(ssl_options=ssl_options)
        self._data_callback = data_callback
        self._channels = dict()
        self._router = router

    def handle_stream(self, stream, address):
        logger.debug("Creating channel for new connection from client {}".format(address))
        self._channels[address] = self._create_channel(stream, address)

    def on_remove_channel(self, client_address):
        channel = self._channels[client_address]
        del self._channels[client_address]
        del channel

    def _create_channel(self, stream, client_address):
        raise NotImplementedError()

    def on_request_complete(self, request: Request, response: Response):
        pass

    @property
    def router(self):
        return self._router


class RecordProxy(ProxyBase):
    def __init__(self, data_callback, router, ssl_options=None):
        super().__init__(data_callback, router, ssl_options=ssl_options)

    def _create_channel(self, stream, client_address):
        return Channel(self, stream, client_address, self.router)

    def on_request_complete(self, request: Request, response: Response):
        self._data_callback.log(request, response)



import asyncio
import json
from meeshkan.schemabuilder.update_mode import UpdateMode
import click
from typing import Sequence, cast

from .config import setup
from .logger import get as getLogger
from .schemabuilder.builder import BASE_SCHEMA, build_schema_async
from .convert.pcap import convert_pcap
from .schemabuilder.schema import validate_openapi_object
from .sinks import AbstractSink, FileSystemSink
from .sources import AbstractSource, KafkaSource, FileSource
from .sources.kafka import KafkaProcessorConfig
from openapi_typed import OpenAPIObject
from yaml import safe_load
from .meeshkan_types import *
from .server.commands import record, mock


LOGGER = getLogger(__name__)


def log(*args):
    LOGGER.debug(*args)


@click.group()
def cli():
    """
    Meeshkan CLI.
    """
    setup()  # Ensure setup is done before invoking the CLI.


async def write_to_sink(result_stream: BuildResultStream, sinks: Sequence[AbstractSink]):
    try:
        async for result in result_stream:
            for sink in sinks:
                sink.push(result)
    finally:
        LOGGER.debug("Flushing all sinks")
        for sink in sinks:
            sink.flush()


def run_from_source(source: AbstractSource, mode: UpdateMode, starting_spec: OpenAPIObject, sinks: Sequence[AbstractSink]):
    loop = asyncio.get_event_loop()

    async def run(loop):
        source_stream, source_task = await source.start(loop)
        result_stream = build_schema_async(source_stream, mode, starting_spec)
        sink_task = loop.create_task(write_to_sink(result_stream, sinks))

        if source_task is not None:
            # TODO Kafka source behaves a bit bad at SIG_INT: it does not raise but returns,
            # leaving the sink task to wait for new input that never arrives. Therefore
            # one cannot `await sink_task` here, at least not without timeout
            await source_task
            try:
                await asyncio.wait_for(sink_task, timeout=3.0)
            except asyncio.TimeoutError:
                LOGGER.warn("Timeout on waiting for sink task to finish.")
        else:
            await sink_task

    try:
        loop.run_until_complete(run(loop))
    finally:
        LOGGER.info("Shutting down source")
        source.shutdown()
        loop.close()


@click.command()
@click.option("-i", "--input-file", required=False, type=click.File('rb'), help="Input file. Use dash '-' to read from stdin.")
@click.option("-o", "--out", required=True, default='out', type=click.Path(exists=False, file_okay=False, writable=True, readable=True), help="Output directory. If the directory does not exist, it is created if the parent directory exists.")
@click.option("-a", "--initial-openapi-spec", required=False, type=click.File('rb'), help="Initial OpenAPI spec.")
@click.option("-m", "--mode", required=False, default='gen', type=str , help="Spec building mode (replay, gen, mixed).")
@click.option("--source", required=False, default='file', type=str, help="Source to read recordings from. For example, 'kafka'")
@click.option("--sink", required=False, type=str,  help="Sink where to write results.")
def build(input_file, out, initial_openapi_spec, mode, source, sink):
    """
    Build OpenAPI schema from HTTP exchanges.
    """

    if input_file is not None and source != "file":
        raise Exception("Only specify input-file for --source file")

    sinks = [FileSystemSink(out)]

    if source == 'kafka':
        # TODO Kafka configuration
        source = KafkaSource(config=KafkaProcessorConfig(
            broker="localhost:9092",
            topic="express_recordings"))
    elif source == 'file':
        if input_file is None:
            raise Exception("Option --input-file for source 'file' required.")
        source = FileSource(input_file)
    else:
        raise Exception("Unknown source {}".format(source))

    openapi_spec: OpenAPIObject = BASE_SCHEMA

    if initial_openapi_spec is not None:
        try: 
            maybe_openapi = cast(OpenAPIObject, safe_load(initial_openapi_spec.read())) 
            # will raise if not an API spec 
            validate_openapi_object(maybe_openapi) 
            openapi_spec = maybe_openapi
        except: pass # just use the initial schema
    run_from_source(source, UpdateMode.GEN if mode == 'gen' else UpdateMode.REPLAY if mode == 'replay' else UpdateMode.MIXED, openapi_spec, sinks=sinks)


@click.command()
@click.option("-i", "--input-file", required=True, type=click.Path(exists=True, file_okay=True, dir_okay=False, readable=True, allow_dash=True), help="Path to a packet capture file.")
@click.option("-o", "--out", required=False, default='recordings.jsonl', type=click.Path(exists=False, file_okay=True, dir_okay=False, writable=True, allow_dash=True), help="Output file.")
def convert(input_file, out):
    """
    Convert recordings from PCAP to JSONL format.
    """

    if not input_file.endswith('.pcap'):
        raise ValueError(
            'Only .pcap files are accepted as input. Got: {}'.format(input_file))

    request_response_pairs = convert_pcap(input_file)

    log("Writing to: %s", out)

    counter = 0
    with open(out, 'w') as f:
        for reqres in request_response_pairs:
            f.write((json.dumps(reqres) + "\n"))
            counter += 1

    log("Wrote %d lines.", counter)


cli.add_command(build)  # type: ignore
cli.add_command(convert)  # type: ignore
cli.add_command(record) # type: ignore
cli.add_command(mock) # type: ignore

if __name__ == '__main__':
    cli()



from meeshkan.schemabuilder.update_mode import UpdateMode
from meeshkan.schemabuilder.param import ParamBuilder, SchemaParameters
from http_types import RequestBuilder
from openapi_typed import Parameter
from typing import cast
from hamcrest import *

req = RequestBuilder.from_url(
    "https://petstore.swagger.io/v1/pets?id=32&car=ferrari")


def test_build_new_query():
    query = req['query']
    schema_query = ParamBuilder('query').build(query, UpdateMode.GEN)
    assert_that(schema_query, has_length(2))

    query_parameter = get_query_parameter("id", schema_query)

    assert_that(query_parameter, has_entry('name', 'id'))
    assert_that(query_parameter, has_entry('in', 'query'))

    assert_that(query_parameter, has_key("schema"))

    query_schema = query_parameter['schema']

    assert_that(query_schema, has_entry('type', 'string'))


required_query_parameter = Parameter(name='key',
                                     required=True,
                                     schema={'type': 'string'}, **{'in': 'query'})


def get_query_parameter(name: str, parameters: SchemaParameters) -> Parameter:
    matching = [param for param in parameters if param['name'] == name][0]
    return cast(Parameter, matching)


def test_update_query_should_update_required():
    query = req['query']
    updated_query = ParamBuilder('query').update(query, UpdateMode.GEN, [required_query_parameter])
    assert_that(updated_query, has_length(3))

    updated_for_key = get_query_parameter("key", updated_query)
    assert_that(updated_for_key, has_entry('required', False))



import collections
import os
import typing
import urllib.parse

from meeshkan.server.utils.http_utils import split_path

RoutingInfo = collections.namedtuple('RoutingInfo', ['host', 'hostname', 'path', 'scheme', 'port'])


def get_port(scheme: str, port: typing.Optional[int]) -> int:
    if port is None:
        if scheme == 'https':
            return 443
        elif scheme == 'http':
            return 80
        else:
            raise AttributeError('Unsupported scheme {}'.format(scheme))

    return port


class Routing:
    def route(self, path: str, headers: typing.Dict, inbound_scheme='http') -> RoutingInfo:
        raise NotImplementedError()


class PathRouting(Routing):
    def route(self, path: str, headers: typing.Dict, inbound_scheme='http') -> RoutingInfo:
        splits = split_path(path)
        url = urllib.parse.urlsplit('{}//{}'.format(splits[0], splits[1]))
        path = '/' + '/'.join(splits[2:])
        return RoutingInfo(path=path, host=splits[1], hostname=url.hostname, port=get_port(url.scheme, url.port),
                           scheme=url.scheme)


class HeaderRouting(Routing):
    def route(self, path: str, headers: typing.Dict, inbound_scheme='http') -> RoutingInfo:
        url = urllib.parse.urlsplit('//{}'.format(headers['Host']))
        return RoutingInfo(path=path, host=headers['Host'], hostname=url.hostname,
                           port=get_port(inbound_scheme, url.port), scheme=inbound_scheme)


class StaticRouting(Routing):
    def __init__(self, target_address):
        self._url = urllib.parse.urlsplit(target_address)
        self._host = '{}:{}'.format(self._url.hostname, self._url.port) if self._url.port else self._url.hostname

    def route(self, path: str, headers: typing.Dict, inbound_scheme='http') -> RoutingInfo:
        return RoutingInfo(path=path, host=self._host, hostname=self._url.hostname,
                           port=get_port(self._url.scheme, self._url.port), scheme=self._url.scheme)



from meeshkan.gen.generator import match_urls, use_if_header, ref_name, get_path_item_with_method, matches
from openapi_typed import PathItem, OpenAPIObject

def test_match_urls():
    assert ["https://api.foo.com"] == match_urls("https", "api.foo.com", {
      "openapi": "",
      "paths": {},
      "info": { "title": "", "version": "" },
      "servers": [
        { "url": "https://api.foo.commm/v2" },
        { "url": "https://api.foo.com" },
      ],
    })
    assert ["https://api.foo.com/v2/a/b/c"] == match_urls("https", "api.foo.com", {
      "openapi": "",
      "paths": {},
      "info": { "title": "", "version": "" },
      "servers": [
        { "url": "https://api.foo.commm/v2" },
        { "url": "https://api.foo.com/v2/a/b/c" },
      ],
    })
    assert [] == match_urls("https", "api.foo.com/v2", {
      "openapi": "",
      "paths": {},
      "info": { "title": "", "version": "" },
      "servers": [
        { "url": "https://api.foo.commm/v2" },
        { "url": "https://api.foo.cooom/v2" },
      ],
    })

def test_get_path_item_with_method():
  o: PathItem = {
    'get': { 'responses': { '100': { 'description': "hello" } } },
    'post': { 'responses': { '101': { 'description': "hello" } } },
    'delete': { 'responses': { '102': { 'description': "hello" } } },
    'description': "foo",
  }
  assert get_path_item_with_method("get", o) == {
    'get': { 'responses': { '100': { 'description': "hello" } } },
    'description': "foo",
  }
  assert get_path_item_with_method("post", o) == {
    'post': { 'responses': { '101': { 'description': "hello" } } },
    'description': "foo",
  }

def test_matcher():
  bfoo = {
    'parameters': [
      {
        'in': "path",
        'name': "foo",
        'schema': { 'type': "string", 'pattern': "^[abc]+$" },
      },
    ],
  };
  oai: OpenAPIObject = {
    'openapi': "",
    'info': { 'title': "", 'version': "" },
    'paths': {
      "/b/{foo}": bfoo,
    },
  }
  assert matches("/a/b", "/a/b", bfoo, "get", oai)
  assert not matches("/a/", "/a/b", bfoo, "get", oai)
  assert not matches("/a/b", "/a", bfoo, "get", oai)
  assert matches("/a/b/c", "/a/{fewfwef}/c", bfoo, "get", oai)
  assert matches("/b/ccaaca", "/b/{foo}", bfoo, "get", oai)
  assert not matches("/b/ccaacda", "/b/{foo}", bfoo, "get", oai)

def test_ref_name():
  assert "Foo" == ref_name({ '$ref': "#/components/schemas/Foo" })

baseO: OpenAPIObject = {
  'openapi': "hello",
  'info': { 'title': "", 'version': "" },
  'servers': [{ 'url': "https://hello.api.com" }],
  'paths': {},
};

def test_use_if_header():
  assert use_if_header(baseO, { 'name': "foo", 'in': "query" }) is None
  assert use_if_header(baseO, { 'name': "foo", 'in': "header" }) == ("foo", { 'type': "string" })
  assert use_if_header(baseO, {
      'name': "foo",
      'in': "header",
      'schema': { 'type': "number" },
  }) == ("foo", { 'type': "number" })
  assert use_if_header(
      {
        **baseO,
        'components': { 'schemas': { 'Foo': { 'type': "boolean" } } },
      },
      {
        'name': "foo",
        'in': "header",
        'schema': { '$ref': "#/components/schemas/Foo" },
      },
    ) == ("foo", { 'type': "boolean" })



import copy
from .update_mode import UpdateMode
from genson import SchemaBuilder  # type: ignore


def _to_openapi_compatible(schema):
    """OpenAPI does not strictly use JSON schema so convert to the OpenAPI subset.

    https://swagger.io/docs/specification/data-models/keywords/

    Arguments:
        schema -- Schema

    Returns:
        Modified schema
    """
    schema = copy.deepcopy(schema)
    if '$schema' in schema:
        del schema['$schema']
    return schema

def to_const(obj):
    if type(obj) == type(''):
        return { 'type': 'string', 'enum': [obj] }
    if type(obj) == type(0):
        return { 'type': 'integer', 'enum': [obj] }
    if type(obj) == type(0.0):
        return { 'type': 'number', 'enum': [obj] }
    if obj is None:
        return { 'type': 'null' }
    if type(obj) == type(True):
        return { 'type': 'boolean', 'enum': [obj] }
    if type(obj) == type([]):
        return { 'type': 'array', 'items': [to_const(x) for x in obj] }
    return {
        'type': 'object',
        'properties': { k: to_const(v) for k, v in obj.items() },
        'required': [x for x in obj.keys()]
    }


def to_json_schema(obj, mode: UpdateMode, schema=None):
    """Create JSON schema based on an object.

    Arguments:
        obj {dict} -- Dictionary object

    Keyword Arguments:
        schema {dict} -- Existing schema if exists. (default: {None})

    Returns:
        [dict] -- New or updated schema.
    """
    if mode == UpdateMode.GEN:
        builder = SchemaBuilder()
        if schema is not None:
            builder.add_schema(schema)
        builder.add_object(obj)
        schema = builder.to_schema()
        return schema
    elif schema is None:
        return { 'oneOf': [to_const(obj)] }
    else:
        return { 'oneOf': [to_const(obj), schema]}



def to_openapi_json_schema(obj, mode: UpdateMode, schema=None):
    """Create OpenAPI-compatible JSON schema from object.

    https://swagger.io/docs/specification/data-models/keywords/

    Arguments:
        obj {[type]} -- [description]

    Keyword Arguments:
        schema {[type]} -- [description] (default: {None})

    Returns:
        [type] -- [description]
    """
    schema = to_json_schema(obj, mode, schema)
    openapi_compatible_schema = _to_openapi_compatible(schema)
    return openapi_compatible_schema



import collections
import json
import logging
import os
import socket
import typing
from enum import Enum
from urllib import parse
from urllib.parse import urlsplit
from http_types import Request, Response, HttpMethod

from meeshkan.server.utils.routing import Routing
from .proxy_callback import ProxyCallback
from ..utils.http_utils import split_path, response_from_bytes
from tornado.iostream import IOStream, SSLIOStream, StreamClosedError



logger = logging.getLogger(__name__)


class ConnectionState(Enum):
    CLOSED = 0
    CONNECTING = 1
    CONNECTED = 2
    WRITING = 3


RequestInfo = collections.namedtuple('RequestInfo', ['data', 'scheme', 'target_host', 'target_port'])


class StreamWrapper:
    def __init__(self, stream: IOStream, close_callback: typing.Callable, queue_on_connecting: bool):
        self._queue = []
        self._stream = stream
        self._state = ConnectionState.CONNECTING

        self._stream.set_nodelay(True)
        self._stream.set_close_callback(close_callback)
        self._queue_on_connecting = queue_on_connecting

    def on_connect(self, read_callback: typing.Callable):
        self._state = ConnectionState.CONNECTED

        try:
            self._stream.read_until_close(lambda x: None, read_callback)
        except StreamClosedError:
            self._state = ConnectionState.CLOSED

    def write(self, data: bytes):
        if self._state not in {ConnectionState.CONNECTED, ConnectionState.WRITING}:
            if self._queue_on_connecting:
                self._queue.append(data)
            return

        if self._state != ConnectionState.WRITING:
            self._do_write(data)
        else:
            self._queue.append(data)

    def _do_write(self, data):
        if not data:
            self._state = ConnectionState.CLOSED
            try:
                self._stream.close()
            except StreamClosedError:
                pass
        else:
            self._client_state = ConnectionState.WRITING
            try:
                self._stream.write(data, callback=self._on_done_write())
            except StreamClosedError:
                logger.warning('Server stream closed unexpectedly')
                self._client_state = ConnectionState.CLOSED

    def _on_done_write(self):
        if self._queue:
            self._do_write(self._queue.pop(0))
            return
        self._client_state = ConnectionState.CONNECTED

    @property
    def stream(self):
        return self._stream

    @property
    def queue(self):
        return self._queue

    @property
    def state(self):
        return self._state

    def close(self):
        self._state = ConnectionState.CLOSED


class Channel:
    def __init__(self, proxy_callback: ProxyCallback, stream: IOStream, client_address: tuple, router: Routing):
        self._proxy_calback = proxy_callback
        self._router = router

        self._client_stream = StreamWrapper(stream, self.on_client_close, False)
        self._client_stream.on_connect(self.on_client_read)

        self._server_stream = None
        self._client_address = client_address
        self._request = None
        self._response = list()
        self._content_length = None
        self._recieved_bytes = 0

    def on_request(self, data: bytes) -> RequestInfo:
        req_lines = data.decode('utf-8').split('\r\n')
        method, fullpath, protocol = req_lines[0].split(' ')
        parsed_fullpath = parse.urlparse(fullpath)
        query = parse.parse_qs(parsed_fullpath.query)

        headers = {}
        host_line: int = 0
        body_start: int = 0
        last_line: int = 0
        for line in req_lines[1:]:
            last_line += 1
            if not line:
                break
            else:
                header, value = line.split(': ')
                if header == 'Host':
                    host_line = last_line
                headers[header] = value


        body_start = last_line + 1
        body = []
        # TODO: @Nikolay, is cur_id correct?
        for body_line in range(body_start, len(req_lines)):
            if req_lines[body_line]:
                body.append(req_lines[body_line])

        body = '\r\n'.join(body)

        route_info = self._router.route(parsed_fullpath.path, headers)
        fullpath = "{}?{}".format(route_info.path, parsed_fullpath.query) if query else route_info.path

        req_lines[0] = ' '.join((method, fullpath, protocol))
        req_lines[host_line] = 'Host: {}'.format(route_info.host)
        headers['Host'] = route_info.host

        data = '\r\n'.join(req_lines).encode('utf-8')

        # ignoring type due to this error
        '''
          46:34 - error: Argument of type 'str' cannot be assigned to parameter 'method' of type 'Literal['connect', 'head', 'trace', 'options', 'delete', 'patch', 'post', 'put', 'get']'
          'str' cannot be assigned to 'Literal['connect']'
          'str' cannot be assigned to 'Literal['head']'
          'str' cannot be assigned to 'Literal['trace']'
          'str' cannot be assigned to 'Literal['options']'
          'str' cannot be assigned to 'Literal['delete']'
        '''
        self._request = Request(method=typing.cast(HttpMethod, method.lower()), # type: ignore
                                host=route_info.host,
                                path=fullpath,
                                pathname=route_info.path,
                                protocol=route_info.scheme,
                                query=query,
                                body=body,
                                bodyAsJson=json.loads(body) if body else {},
                                headers=headers)

        return RequestInfo(data=data, scheme=route_info.scheme, target_host=route_info.hostname, target_port=route_info.port)

    def on_response_chunk(self, data: bytes):
        if len(self._response) == 0:
            resp_lines = data.split(b'\r\n')
            for line in resp_lines[1:]:
                res = line.split(b': ')
                if len(res) == 2:
                    header, value = res[0].decode('utf-8'), res[1].decode('utf-8')
                    if header == 'Content-Length':
                        self._content_length = int(value)
                        break

        self._response.append(data)
        self._recieved_bytes += len(data)

        logger.debug('[%s] Received response chunk %d bytes out of %d', self._client_address,
                     self._recieved_bytes, self._content_length or 0)

    def flush(self, check_length=True):
        if len(self._response) > 0 and self._request is not None:
            if not check_length or self._content_length is None or self._recieved_bytes >= self._content_length:
                logger.debug('[%s] Flushing request, received %d bytes out of %d', self._client_address,
                             self._recieved_bytes, self._content_length or 0)

                resp = b''.join(self._response)
                resp = response_from_bytes(resp)
                body = resp.data.decode('utf-8')

                resp = Response(statusCode=resp.status, body=body, bodyAsJson=json.loads(body),
                                headers=dict(resp.getheaders()))

                self._proxy_calback.on_request_complete(self._request, resp)

                self._request = None
                self._response = []
                self._content_length = None
                self._recieved_bytes = 0
        else:
            logger.debug('[%s] Trying to flush incomplete data, request "%s", response chunks %d', self._client_address,
                         str(self._request), len(self._response))

    def on_server_read(self, data):
        self.on_response_chunk(data)
        self._client_stream.write(data)

    def on_client_read(self, data):
        self.flush()

        request_info = self.on_request(data)

        logger.debug('[%s] Got request for %s://%s:%s', self._client_address, request_info.scheme,
                     request_info.target_host, request_info.target_port)

        if self._server_stream is None:
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM, 0)
            stream = SSLIOStream(s, ssl_options={}) if request_info.scheme == 'https' else IOStream(s)
            stream.connect((request_info.target_host, request_info.target_port), self.on_server_connect)
            self._server_stream = StreamWrapper(stream, self.on_server_close, True)

        self._server_stream.write(request_info.data)

    def on_server_connect(self):
        self._server_stream.on_connect(self.on_server_read)
        if self._server_stream.queue:
            self._server_stream.write(self._server_stream.queue.pop(0))

    def remove_channel(self):
        self.flush(check_length=False)
        self._proxy_calback.on_remove_channel(self._client_address)

    def on_server_close(self):
        self._server_stream.close()
        if self._client_stream.state == ConnectionState.CLOSED:
            self.remove_channel()
        else:
            self._client_stream.write(b'')

    def on_client_close(self):
        self._client_stream.close()
        if self._server_stream is None or self._server_stream.state == ConnectionState.CLOSED:
            self.remove_channel()
        else:
            self._server_stream.write(b'')

    @property
    def server_state(self):
        return ConnectionState.CLOSED if self._server_stream is None else self._server_stream.state

    @property
    def client_state(self):
        return self._client_stream.state





import logging
import logging.config
from typing import List
from .config import setup

# Do not expose anything by default (internal module)
__all__ = []  # type: List[str]

# Loggers cached by name
_loggers = {}


def get(name):
    setup()  # Setup needed before creating any loggers
    if name in _loggers:
        return _loggers[name]
    logger = logging.getLogger(name)
    _loggers[name] = logger
    return logger



"""OpenAPI schema validation operations.
"""

from jsonschema import validate
import json
import os
from openapi_typed import OpenAPIObject

_schema = None


def _openapi_json_schema():
    global _schema
    if _schema is not None:
        return _schema
    path = os.path.dirname(os.path.realpath(__file__))
    with open(path + "/schema.json") as f:
        _schema = json.loads(f.read())

    return _schema


def validate_openapi_object(instance: OpenAPIObject):
    return validate(instance=instance, schema=_openapi_json_schema())



from meeshkan.gen.generator import matcher
from openapi_typed import OpenAPIObject
from typing import Dict
import json

store: Dict[str, OpenAPIObject] = {
  'foo': {
    'openapi': "",
    'servers': [{ 'url': "https://api.foo.com" }],
    'info': { 'title': "", 'version': "" },
    'paths': {
      "/user": {
        'get': { 'responses': { '200': { 'description': "userget" } } },
        'post': { 'responses': { '200': { 'description': "userpost" } } },
        'description': "",
      },
      "/user/{id}": {
        'parameters': [
          {
            'name': "id",
            'in': "path",
            'required': True,
            'schema': {
              'type': "integer",
            },
          },
        ],
        'get': { 'responses': { '200': { 'description': "useridget" } } },
        'post': { 'responses': { '201': { 'description': "useridpost" } } },
      },
      "/user/{id}/name": {
        'get': { 'responses': { '200': { 'description': "useridnameget" } } },
        'post': { 'responses': { '201': { 'description': "useridnamepost" } } },
        'description': "",
      },
    },
  },
  'bar': {
    'openapi': "",
    'servers': [{ 'url': "https://api.bar.com" }],
    'info': { 'title': "", 'version': "" },
    'paths': {
      "/guest": {
        'get': { 'responses': { '200': { 'description': "guestget" } } },
        'post': { 'responses': { '200': { 'description': "guestpost" } } },
        'description': "",
      },
      "/guest/{id}": {
        'get': { 'responses': { '200': { 'description': "guestidget" } } },
        'post': { 'responses': { '201': { 'description': "guestidpost" } } },
        'description': "",
      },
      "/guest/{id}/name": {
        'get': { 'responses': { '200': { 'description': "guestidnameget" } } },
        'post': { 'responses': { '201': { 'description': "guestidnamepost" } } },
        'description': "",
      },
    },
  },
  'baz': {
    'openapi': "",
    'servers': [{ 'url': "https://api.baz.com" }],
    'info': { 'title': "", 'version': "" },
    'paths': {
      "/guest": {
        'parameters': [
          {
            'in': "query",
            'required': True,
            'name': "hello",
            'schema': { 'type': "string", 'pattern': "^[0-9]+$" },
          },
        ],
        'get': { 'responses': { '200': { 'description': "guestget" } } },
        'post': {
          'parameters': [
            {
              'in': "query",
              'required': True,
              'name': "thinking",
              'schema': { 'type': "string" },
            },
          ],
          'responses': { '200': { 'description': "guestpost" } },
        },
        'description': "",
      },
      "/guest/{id}": {
        'parameters': [
          {
            'in': "query",
            'required': True,
            'name': "a",
            'schema': { 'type': "string" },
          },
          {
            'in': "query",
            'required': True,
            'name': "b",
            'schema': { 'type': "string" },
          },
          {
            'in': "query",
            'name': "c",
            'schema': { 'type': "string" },
          },
        ],
        'get': { 'responses': { '200': { 'description': "guestidget" } } },
        'post': {
          'parameters': [
            {
              'in': "header",
              'required': True,
              'name': "zz",
              'schema': { 'type': "string" },
            },
            {
              'in': "query",
              'required': True,
              'name': "zzz",
              'schema': { 'type': "string" },
            },
            {
              'in': "query",
              'name': "c",
              'schema': { 'type': "string" },
            },
          ],
          'responses': { '201': { 'description': "guestidpost" } },
        },
        'description': "",
      },
      "/guest/{id}/name": {
        'get': { 'responses': { '200': { 'description': "guestidnameget" } } },
        'post': {
          'requestBody': {
            'content': {
              "application/json": {
                'schema': {
                  'type': "object",
                  'required': ["age"],
                  'properties': { 'age': { 'type': "integer" } },
                },
              },
            },
          },
          'responses': { '201': { 'description': "guestidnamepost" } },
        },
        'description': "",
      },
    },
  },
}

def test_matcher_1():
  assert matcher(
      {
        'headers': {},
        'host': "api.foo.com",
        'path': "/user",
        'pathname': "/user",
        'protocol': "https",
        'method': "get",
        'query': {},
      },
      store
    ) == {
    'foo': {
      'openapi': "",
      'servers': [{ 'url': "https://api.foo.com" }],
      'info': { 'title': "", 'version': "" },
      'paths': {
        "/user": {
          'get': { 'responses': { '200': { 'description': "userget" } } },
          'description': "",
        },
      },
    },
  }

def test_matcher_2():
    matcher(
      {
        'headers': {},
        'host': "api.bar.com",
        'path': "/guest/{id}",
        'pathname': "/guest/{id}",
        'protocol': "https",
        'method': "post",
        'query': {},
      },
      store
    ) == {
        'bar': {
            'openapi': "",
            'servers': [{ 'url': "https://api.bar.com" }],
            'info': { 'title': "", 'version': "" },
            'paths': {
                "/guest/{id}": {
                'post': { 'responses': { '201': { 'description': "guestidpost" } } },
                'description': "",
                },
            }
        }
    }

def test_matcher_3():
    assert matcher(
      {
        'headers': {},
        'host': "api.foo.com",
        'path': "/users", # incorrect, should be user
        'pathname': "/users", # incorrect, should be user
        'protocol': "https",
        'method': "get",
        'query': {},
      },
      store
    ) == {
    'foo': {
      'openapi': "",
      'servers': [{ 'url': "https://api.foo.com" }],
      'info': { 'title': "", 'version': "" },
      'paths': {},
    }
  }

def test_matcher_4():
    assert matcher(
      {
        'headers': {},
        'host': "api.foo.com",
        'path': "/user/55", # correctly parses number
        'pathname': "/user/55", # correcly parses number
        'protocol': "https",
        'method': "get",
        'query': {},
      },
      store
    ) == {
    'foo': {
      'openapi': "",
      'servers': [{ 'url': "https://api.foo.com" }],
      'info': { 'title': "", 'version': "" },
      'paths': {
        "/user/{id}": {
          'parameters': [
            {
              'name': "id",
              'in': "path",
              'required': True,
              'schema': {
                'type': "integer",
              },
            },
          ],
          'get': { 'responses': { '200': { 'description': "useridget" } } },
        },
      },
    },
  }

def test_matcher_5():
    assert matcher(
      {
        'headers': {},
        'host': "api.foo.com",
        'path': "/user/fdsfsfwef", # correctly rejects non-number
        'pathname': "/user/fdsfsfwef", # correcly rejects non-number
        'protocol': "https",
        'method': "get",
        'query': {},
      },
      store
    ) == {
    'foo': {
      'openapi': "",
      'servers': [{ 'url': "https://api.foo.com" }],
      'info': { 'title': "", 'version': "" },
      'paths': {},
    },
  }

def test_matcher_6():
    assert matcher(
      {
        'headers': {},
        'host': "api.foo.com",
        'path': "/user",
        'pathname': "/user",
        'protocol': "https",
        'method': "delete", # does not exist
        'query': {},
      },
      store,
    ) == {
    'foo': {
      'openapi': "",
      'servers': [{ 'url': "https://api.foo.com" }],
      'info': { 'title': "", 'version': "" },
      'paths': {
        "/user": {
          'description': "",
        },
      },
    },
  }

def test_matcher_7():
    assert matcher(
      {
        'headers': {},
        'host': "api.foo.commmm", # does not exist
        'path': "/user",
        'pathname': "/user",
        'protocol': "https",
        'method': "get",
        'query': {},
      },
      store,
    ) == {}

def test_matcher_8():
    assert matcher(
      {
        'headers': {},
        'host': "api.baz.com",
        'path': "/guest",
        'pathname': "/guest",
        'protocol': "https",
        'method': "get",
        'query': { 'hello': "0" }, # query param must be integer
      },
      store,
    )['baz']['paths']["/guest"]['get'] == store['baz']['paths']["/guest"]['get']

def test_matcher_9():
    assert 'get' not in matcher(
      {
        'headers': {},
        'host': "api.baz.com",
        'path': "/guest",
        'pathname': "/guest",
        'protocol': "https",
        'method': "get",
        'query': { 'hello': "b" }, # query param intentionally off as string
      },
      store,
    )['baz']['paths']["/guest"]

def test_matcher_10():
  assert 'get' not in matcher(
      {
        'headers': {},
        'host': "api.baz.com",
        'path': "/guest",
        'pathname': "/guest",
        'protocol': "https",
        'method': "get",
        'query': {}, # as query is required, this should fail
      },
      store,
    )['baz']['paths']["/guest"]

def test_matcher_11():
    assert matcher(
      {
        'headers': {},
        'host': "api.baz.com",
        'path': "/guest/3/name",
        'pathname': "/guest/3/name",
        'protocol': "https",
        'method': "post",
        'query': {},
        'bodyAsJson': { 'age': 42 },
        'body': json.dumps({ 'age': 42 }),
      },
      store,
    )['baz']['paths']["/guest/{id}/name"]['post'] == store['baz']['paths']["/guest/{id}/name"]['post']

def test_matcher_12():
    assert 'post' not in matcher(
      {
        'headers': {},
        'host': "api.baz.com",
        'path': "/guest/3/name",
        'pathname': "/guest/3/name",
        'protocol': "https",
        'method': "post",
        'query': {},
        'body': json.dumps({ 'age': "42" }),
        'bodyAsJson': { 'age': "42" }, # wrong type, as 42 is string
      },
      store,
    )['baz']['paths']["/guest/{id}/name"]

def test_matcher_13():
    assert matcher(
      {
        'headers': { 'zz': "top" },
        'host': "api.baz.com",
        'path': "/guest/4",
        'pathname': "/guest/4",
        'protocol': "https",
        'method': "post",
        'query': { 'zzz': "aaa", 'a': "foo", 'b': "baz" },
      },
      store,
    )['baz']['paths']["/guest/{id}"]['post'] == store['baz']['paths']["/guest/{id}"]['post']

def test_matcher_14():
    assert 'post' not in matcher(
      {
        'headers': {}, # no header will lead to undefined
        'host': "api.baz.com",
        'path': "/guest/4",
        'pathname': "/guest/4",
        'protocol': "https",
        'method': "post",
        'query': { 'zzz': "aaa", 'a': "foo", 'b': "baz" },
      },
      store,
    )['baz']['paths']["/guest/{id}"]



import json
import logging
from typing import cast
from urllib import parse

from http_types import Request, HttpMethod
from tornado.web import RequestHandler

from meeshkan.server.server.callbacks import callback_manager

logger = logging.getLogger(__name__)


class MockServerView(RequestHandler):
    SUPPORTED_METHODS = ["GET", "POST", "HEAD", "DELETE", "PATCH", "PUT", "OPTIONS"]

    def set_default_headers(self):
        self.set_header("Content-Type", 'application/json; charset="utf-8"')

    def get(self, **kwargs):
        self._serve()

    def post(self):
        self._serve()

    def head(self, **kwargs):
        self._serve()

    def delete(self, **kwargs):
        self._serve()

    def patch(self, **kwargs):
        self._serve()

    def put(self, **kwargs):
        self._serve()

    def options(self, **kwargs):
        self._serve()

    def _serve(self):
        headers = {k: v for k, v in self.request.headers.get_all()}
        route_info = self.application.router.route(self.request.path, headers)
        headers['Host'] = route_info.host

        query = parse.parse_qs(self.request.query)
        fullpath = "{}?{}".format(route_info.path, self.request.query) if query else route_info.path

        # ignoring type due to this error
        '''
          46:34 - error: Argument of type 'str' cannot be assigned to parameter 'method' of type 'Literal['connect', 'head', 'trace', 'options', 'delete', 'patch', 'post', 'put', 'get']'
          'str' cannot be assigned to 'Literal['connect']'
          'str' cannot be assigned to 'Literal['head']'
          'str' cannot be assigned to 'Literal['trace']'
          'str' cannot be assigned to 'Literal['options']'
          'str' cannot be assigned to 'Literal['delete']'
        '''
        request = Request(method=cast(HttpMethod, self.request.method.lower()),  # type: ignore
                          host=route_info.host,
                          path=fullpath,
                          pathname=route_info.path,
                          protocol=route_info.scheme,
                          query=query,
                          body=self.request.body,
                          bodyAsJson=self._extract_json_safely(self.request.body),
                          headers=headers)

        logger.debug(request)
        response = callback_manager(request, self.application.response_matcher.get_response(request))
        for header, value in response['headers'].items():
            self.set_header(header, value)

        self.write(response['body'])

    def _extract_json_safely(self, text):
        if text:
            try:
                return json.loads(text)
            except Exception as e:
                pass

        return {}



import os
from http.client import HTTPResponse
from http.server import BaseHTTPRequestHandler
from io import BytesIO
from socket import socket

import urllib3


def split_path(path):
    splits = list()
    while path != '/':
        path, split = os.path.split(path)
        splits.append(split)
    return splits[::-1]


class BytesIOSocket(socket):
    def __init__(self, content):
        self.handle = BytesIO(content)

    def makefile(self, mode):
        return self.handle


def response_from_bytes(data) -> urllib3.HTTPResponse:
    sock = BytesIOSocket(data)

    response = HTTPResponse(sock)
    response.begin()

    return urllib3.HTTPResponse.from_httplib(response)


class HTTPRequest(BaseHTTPRequestHandler):
    def __init__(self, request_text):
        self.rfile = BytesIO(request_text)
        self.raw_requestline = self.rfile.readline()
        self.error_code = self.error_message = None
        self.parse_request()

    def send_error(self, code, message=None, explain=None):
        self.error_code = code
        self.error_message = message



import logging

logger = logging.getLogger(__name__)


class StorageManager:
    def __init__(self):
        self._default = dict()
        self._storages = dict()

    @property
    def default(self):
        return self._default

    def get(self, name):
        return self._storages.get(name)

    def add_storage(self, name):
        self._storages[name] = dict()

    def clear(self):
        self._default.clear()
        for storage in self._storages.values():
            storage.clear()
        logger.debug('Storage cleared')


storage_manager = StorageManager()



from typing import Awaitable
import asyncio
from typing import Callable, Awaitable, AsyncIterable, Tuple
from http_types import HttpExchange
from .schemabuilder.result import BuildResult

HttpExchangeStream = AsyncIterable[HttpExchange]
BuildResultStream = AsyncIterable[BuildResult]

Source = Callable[[asyncio.events.AbstractEventLoop],
                  Tuple[HttpExchangeStream, asyncio.Task]]
Sink = Callable[[BuildResultStream], Awaitable[None]]



import os
import sys
from shutil import rmtree

from setuptools import setup, Command, find_packages, errors

# Package meta-data.
NAME = 'meeshkan'
DESCRIPTION = 'Reverse engineer services with style'
URL = 'http://github.com/meeshkan/meeshkan'
EMAIL = 'dev@meeshkan.com'
AUTHOR = 'Meeshkan Dev Team'
REQUIRES_PYTHON = '>=3.6.0'

here = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = '\n' + f.read()

REQUIRED = [
    'click',
    'deepdiff',
    'lenses',
    'pyyaml',
    'jsonschema',
    'faker',
    'typing-extensions',
    'openapi-typed>=0.1.0',
    'typeguard>=2.7.0',
    'genson',
    'http-types>=0.0.5',
    # kafka
    'faust',
    # server
    'tornado==5.1.1',
    'urllib3==1.24.1'
]

BUNDLES = {
}

# Requirements of all bundles
BUNDLE_REQUIREMENTS = [dep for _, bundle_dep in BUNDLES.items()
                       for dep in bundle_dep]

DEV = BUNDLE_REQUIREMENTS + [
    'pytest',
    'pylint',
    'pytest-tornado',
    'setuptools',
    'twine',
    'wheel',
    'pytest-watch',
    'pytest-testmon',
    'pyhamcrest',
    'pytest-asyncio'
]

VERSION = '0.2.6'

ENTRY_POINTS = ['meeshkan = meeshkan.__main__:cli']

EXTRAS = dict(**BUNDLES, dev=DEV)


class SetupCommand(Command):
    """Base class for setup.py commands with no arguments"""
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print('\033[1m{0}\033[0m'.format(s))

    def rmdir_if_exists(self, directory):
        self.status("Deleting {}".format(directory))
        rmtree(directory, ignore_errors=True)


def build():
    return os.system(
        "{executable} setup.py sdist bdist_wheel --universal".format(executable=sys.executable))


def type_check():
    os.system("pyright --lib")


class BuildDistCommand(SetupCommand):
    """Support setup.py upload."""
    description = "Build the package."

    def run(self):
        self.status("Removing previous builds...")
        self.rmdir_if_exists(os.path.join(here, 'dist'))

        self.status("Building Source and Wheel (universal) distribution...")
        build()
        sys.exit()


class TypeCheckCommand(SetupCommand):
    """Run type-checking."""
    description = "Run type-checking."

    def run(self):
        type_check()
        sys.exit()


class TestCommand(SetupCommand):
    """Support setup.py test."""
    description = "Run local test if they exist"

    def run(self):
        os.system("pytest")
        sys.exit()


class UploadCommand(SetupCommand):
    """Support setup.py upload."""
    description = "Build and publish the package."

    def run(self):

        self.status("Removing previous builds...")
        self.rmdir_if_exists(os.path.join(here, 'dist'))

        self.status("Building Source and Wheel (universal) distribution...")
        exit_code = build()
        if exit_code != 0:
            raise errors.DistutilsError("Build failed.")
        self.status("Uploading the package to PyPI via Twine...")
        os.system("twine upload dist/*")

        self.status("Pushing git tags...")
        os.system("git tag v{about}".format(about=VERSION))
        os.system("git push --tags")

        sys.exit()


setup(name=NAME,
      version=VERSION,
      description=DESCRIPTION,
      long_description=long_description,
      long_description_content_type='text/markdown',
      url=URL,
      author=AUTHOR,
      author_email=EMAIL,
      python_requires=REQUIRES_PYTHON,
      license='MIT',
      packages=find_packages(exclude=('tests',)),
      include_package_data=True,
      install_requires=REQUIRED,
      extras_require=EXTRAS,
      classifiers=[
          'Programming Language :: Python :: 3',
          'License :: OSI Approved :: MIT License',
          'Operating System :: OS Independent',
      ],
      zip_safe=False,
      entry_points={'console_scripts': ENTRY_POINTS},
      cmdclass={'dist': BuildDistCommand,
                'upload': UploadCommand,
                'test': TestCommand,
                'typecheck': TypeCheckCommand}
      )



import json
import logging
import os
from urllib import parse
from urllib.parse import urlencode

from http_types import RequestBuilder, Request

from ..server.storage import storage_manager
from tornado.web import RequestHandler

logger = logging.getLogger(__name__)


class StorageView(RequestHandler):
    SUPPORTED_METHODS = ["DELETE"]

    def set_default_headers(self):
        self.set_header("Content-Type", 'application/json; charset="utf-8"')

    def delete(self, **kwargs):
        storage_manager.clear()






from meeshkan.gen.generator import faker, valid_schema
import jsonschema

def test_faker_1():
    schema = {
  "type": "array",
  "items": {
    "type": "object",
    "required": [
      "foo",
      "baz"
    ],
    "properties": {
      "foo": {
        "type": "number"
      },
      "bar": {
        "type": "string"
      },
      "baz": {
        "type": "string"
      }
    }
  }
}
    res = faker(schema, schema, 0)
    assert valid_schema(res, schema)

def test_faker_2():
    schema = {
  "$id": "https://example.com/person.schema.json",
  "$schema": "http://json-schema.org/draft-07/schema#",
  "title": "Person",
  "type": "object",
  "properties": {
    "firstName": {
      "type": "string",
      "description": "The person's first name."
    },
    "lastName": {
      "type": "string",
      "description": "The person's last name."
    },
    "age": {
      "description": "Age in years which must be equal to or greater than zero.",
      "type": "integer",
      "minimum": 0
    }
  }
}
    res = faker(schema, schema, 0)
    assert valid_schema(res, schema)

def test_faker_3():
    schema = {
  "$id": "https://example.com/geographical-location.schema.json",
  "$schema": "http://json-schema.org/draft-07/schema#",
  "title": "Longitude and Latitude Values",
  "description": "A geographical coordinate.",
  "required": [ "latitude", "longitude" ],
  "type": "object",
  "properties": {
    "latitude": {
      "type": "number",
      "minimum": -90,
      "maximum": 90
    },
    "longitude": {
      "type": "number",
      "minimum": -180,
      "maximum": 180
    }
  }
}
    res = faker(schema, schema, 0)
    assert valid_schema(res, schema)

def test_faker_4():
    schema = {
  "$id": "https://example.com/arrays.schema.json",
  "$schema": "http://json-schema.org/draft-07/schema#",
  "description": "A representation of a person, company, organization, or place",
  "type": "object",
  "properties": {
    "fruits": {
      "type": "array",
      "items": {
        "type": "string"
      }
    },
    "vegetables": {
      "type": "array",
      "items": { "$ref": "#/definitions/veggie" }
    }
  },
  "definitions": {
    "veggie": {
      "type": "object",
      "required": [ "veggieName", "veggieLike" ],
      "properties": {
        "veggieName": {
          "type": "string",
          "description": "The name of the vegetable."
        },
        "veggieLike": {
          "type": "boolean",
          "description": "Do I like this vegetable?"
        }
      }
    }
  }
}
    res = faker(schema, schema, 0)
    assert valid_schema(res, schema)



"""Code for converting pcap to HttpExchange JSON format."""

from sys import argv
from http_types import HttpExchange
from typing import List, Optional, Any, Iterator
from http_types.types import HttpExchange
from typing import Iterator
from pathlib import Path
import csv
import json
import shutil
from typing import List, Optional, Dict, Union
from http_types import Request, Response, HttpExchange
import re
import subprocess
from ..logger import get as getLogger

LOGGER = getLogger(__name__)
TSHARK_EXECUTABLE = 'tshark'


def _request_from_tshark(obj) -> Request:
    method = obj["http.request.method"].lower()
    uri = obj['http.request.uri']
    req_line = obj['http.request.line']
    headers = _parse_headers(header_line=req_line)
    body = obj['http.file_data']
    protocol = obj['_ws.col.Protocol'].lower()
    # TODO Parse
    # query = parse_query(obj['http.request.uri.query'])
    query = {}
    # TODO Handle path correctly
    path = uri
    pathname = path
    # TODO
    
    host = headers['x-forwarded-host'] if 'x-forwarded-host' in headers else ''

    if isinstance(host, list):
        raise AssertionError("Did not expect to get list as x-forwarded-host")

    return Request(
        body=body,
        protocol=protocol,
        method=method,
        headers=headers,
        query=query,
        path=path,
        pathname=pathname,
        host=host)


def _response_from_tshark(obj) -> Response:
    status_code = int(obj['http.response.code'])
    res_line = obj['http.response.line']
    body = obj['http.file_data']  # type: str
    headers = _parse_headers(header_line=res_line)
    return Response(statusCode=status_code, body=body, headers=headers)


def _parse_headers(header_line: str) -> Dict[str, Union[str, List[str]]]:
    # Depends on \\r\\n being a line-break, also removes "\\" from the start of every header
    lines = [cleaned
             for line in header_line.split("\\r\\n")
             for cleaned in (re.sub(r"^\\", "", line),)]

    def line_to_tuple(l: str):
        # dec = l.decode('string_escape')
        if len(l) == 0:
            return None
        split = l.split(":", 1)
        return split[0].strip().lower(), split[1].strip()

    return dict(tupl
                for line in lines
                for tupl in (line_to_tuple(line),)
                if tupl is not None)


# from .meeshkan_types import Request, Response, HttpExchange


def _request_response_parser():
    """Create a parser for parsing tshark objects into HttpExchange.

    Raises:
        Exception: [description]

    Returns:
        [type] -- [description]
    """
    requests = {}

    def parse_request_response_if_res(obj: Any) -> Optional[HttpExchange]:
        """
        Handle one line of tshark dictionary.

        If request, store request to internal dictionary.
        If response, return request-response pair.
        """
        stream_id = obj["tcp.stream"]
        req_method = obj["http.request.method"]

        is_request = True if req_method != "" else False

        if is_request:
            requests[stream_id] = obj
            return None

        # Request corresponding to this response
        request = requests.pop(stream_id)
        response = obj
        req = _request_from_tshark(request)
        res = _response_from_tshark(response)
        req_res = HttpExchange(request=req, response=res)
        return req_res

    return parse_request_response_if_res


def transform_tshark(tshark_csv: Iterator[str]) -> Iterator[HttpExchange]:
    """Convert tshark-produced CSV to an iterator of HttpExchanges.
    
    Arguments:
        tshark_csv {Iterator[str]} -- Lines of CSV produced by tshark.
    
    Returns:
        Iterator[HttpExchange] -- [description]
    """
    parse_request_response = _request_response_parser()
    return (req_res
            for obj in csv.DictReader(tshark_csv, quotechar='\'')
            for req_res in (parse_request_response(obj),)
            if req_res is not None)


def check_tshark_exists():
    if shutil.which(TSHARK_EXECUTABLE) is None:
        raise Exception(
            "Could not find executable {} in PATH.".format(TSHARK_EXECUTABLE))


def convert_pcap(filepath: str) -> Iterator[HttpExchange]:

    check_tshark_exists()

    if filepath == '-':
        raise NotImplementedError("File path '-' detected. Reading from stdin is not implemented yet.")

    resolved_path = Path(filepath).resolve()
    if not resolved_path.is_file():
        raise FileNotFoundError("Not a file: {}".format(filepath))

    tshark_cmd = "tshark -r {} -Y http -E separator=, -E aggregator='\t' -E header=y -E occurrence=a -E quote=s -T fields -e tcp.stream -e _ws.col.Protocol -e http.request.method -e http.request.uri -e http.request.uri.path -e http.request.uri.query -e http.request.full_uri -e http.request.line -e http.transfer_encoding -e http.response.code -e http.response_for.uri -e http.response.line -e http.file_data".format(
        resolved_path)

    # TODO Stream stdout with subprocess.Popen instead of reading the whole output into memory
    tshark_proc = subprocess.run(
        tshark_cmd, shell=True, stdout=subprocess.PIPE)

    if tshark_proc.returncode != 0:
        raise Exception("tshark exited with code {}".format(
            tshark_proc.returncode))

    stdout = tshark_proc.stdout.decode()

    return transform_tshark(iter(stdout.splitlines()))


if __name__ == '__main__':
    filename = argv[1]

    out = convert_pcap(filename)

    for o in out:
        print(json.dumps(o))
